import math

import torch


def closed_transform_2d(t: torch.Tensor) -> torch.Tensor:
    r"""Map :math:`t` in :math:`[0, 1]` to points :math:`(x, y)` on the unit circle.

    Map :math:`t \in [0, 1]` to:
    .. math::
        \theta &= 2 \pi t \\
        x &= \cos \theta \\
        y &= \sin \theta

    Args:
        t: Tensor of values in [0, 1]. Shape :math:`(N, 1)`.

    Returns:
        torch.Tensor: Matrix of coordinates (x, y) on the unit circle. Shape :math:`(N, 2)`.
    """
    pi = t.new_tensor(math.pi)
    theta = 2 * pi * t
    x = torch.cos(theta)
    y = torch.sin(theta)

    return torch.cat([x, y], dim = 1)


def closed_transform_3d(ts: torch.Tensor) -> torch.Tensor:
    r"""Map :math:`(t, s)` in :math:`[0, 1]^2` to points :math:`(x, y, z)` on the unit sphere.

    Map :math:`(t, s) \in [0, 1]^2` to:
    .. math::
        \theta &= 2 \pi t \qquad \phi = \pi s
        x &= \sin \phi \cos \theta \\
        y &= \sin \phi \sin \theta \\
        z &= \cos \phi

    Args:
        ts: Tensor of (t, s) values in [0, 1]. Shape :math:`(N, 2)`.

    Returns:
        torch.Tensor: Matrix of coordinates (x, y, z) on the unit sphere. Shape :math:`(N, 3)`.
    """
    t = ts[:, 0:1]
    s = ts[:, 1:2]

    pi = t.new_tensor(math.pi)
    theta = 2 * pi * t
    phi = pi * s

    sin_phi = torch.sin(phi)
    x = sin_phi * torch.cos(theta)
    y = sin_phi * torch.sin(theta)
    z = torch.cos(phi)

    return torch.cat([x, y, z], dim = 1)


import matplotlib.pyplot as plt
import numpy as np
import torch
from torch import nn


def _n_for_sin_peak_locations(x_peaks: torch.Tensor, eps: float = 1e-2) -> list[int]:
    r"""Convert sine-peak locations to Hicks-Henne exponents using :math:`n = ln(0.5)/ln(x_peak)`.

    Clamp values to avoid log related singularities at 0 and 1.

    Args:
        x_peaks: Basis function peak location. Shape :math:`(N,)`.
        eps: Optional clamp tolerance. Default: 1e-2.
    
    Returns:
        list[int]: Exponents `n_list` corresponding to each element of `x_peaks`.
    """
    # Clamp peak values on the left and right to avoid numerical errors in log computation
    x_peaks = torch.clamp(x_peaks, min = eps, max = 1 - eps)
    half = torch.tensor(0.5, dtype = x_peaks.dtype, device = x_peaks.device)
    n_vals = torch.log(half) / torch.log(x_peaks)

    return n_vals.tolist()


def _polyexp_basis(x: torch.Tensor, n_vals: torch.Tensor, m: torch.Tensor) -> torch.Tensor:
    r"""Polynomial exponential Hicks-Henne basis (Eq. 3 in [1]_).

    For each x value in `x`, n value in `n_vals` and given m, computes:
    .. math::
        y = \frac{\left(x \right)^{n} \left(1 - x \right)}{e^{mx}}
    
    Args:
        x: x-coordinates of points on the shape. Shape :math:`(N,)`.
        n_vals: Exponents n in the numerator polynomial. Shape :math:`(K,)`.
        m: Exponent of the exponential in the denominator. Scalar.
    
    Returns:
        torch.Tensor: Basis matrix. Shape (N, K). Rows = x-samples, Cols = n-values.
    
    References:
        [1] Hicks, Raymond M., and Garret N. Vanderplaats. Application of numerical optimization to
            the design of supercritical airfoils without drag-creep. No. 770440. SAE Technical
            Paper, 1977.
    """
    # Use torch pow function to compute all basis functions at once
    # N - count(x), K - count(n_vals), torch.pow(x, n_vals) returns (N, K)
    # Make x (N, 1), n_vals (1, K) and use broadcasting to produce an output (N, K)
    N = x.shape[0]
    n_vals = n_vals.view(1, -1)
    x = x.view(N, 1)
    num = torch.pow(x, n_vals) * (1.0 - x)
    den = torch.exp(m * x)
    polyexp_basis = num / den
    
    return polyexp_basis


def _sin_basis(x: torch.Tensor, n_vals: torch.Tensor, m: torch.Tensor) -> torch.Tensor:
    r"""Sinusoidal Hicks-Henne basis (Eq. 4 in [1]_).

    For each x value in `x`, n value in `n_vals` and given m, computes:
    .. math::
        y = \sin \left( \pi x^{n} \right)^{m}
    
    Args:
        x: x-coordinates of points on the shape. Shape :math:`(N,)`.
        n_vals: Exponents n of the input to the sine. Shape :math:`(K,)`.
        m: Exponent of the sine output. Scalar.
    
    Returns:
        torch.Tensor: Basis matrix. Shape (N, K). Rows = x-samples, Cols = n-values.
    
    References:
        [1] Hicks, Raymond M., and Garret N. Vanderplaats. Application of numerical optimization to
            the design of supercritical airfoils without drag-creep. No. 770440. SAE Technical
            Paper, 1977.
    """
    # Use torch pow function to compute all basis functions at once
    # N - count(x), K - count(n_vals), torch.pow(x, n_vals) returns (N, K)
    # Make x (N, 1), n_vals (1, K) and use broadcasting to produce an output (N, K)
    N = x.shape[0]
    n_vals = n_vals.view(1, -1)
    x = x.view(N, 1)
    z = torch.pow(x, n_vals)
    sin_basis = torch.pow(torch.sin(torch.pi * z), m)
    
    return sin_basis


class HicksHenne(nn.Module):
    r"""Hicks-Henne bump function representation. As described in [1]_.

    This module implements the Hicks-Henne bump functions as described in [1]_ and popularized by
    [2]_. It supports two families of basis functions: polynomial-exponential (Eq. 3 in [1]_) and
    sinusoidal (Eq. 4 in [1]_). These basis functions are applied separately to the upper and lower
    surfaces via learnable participation coefficients, allowing for flexible shape modifications
    during optimization.

    The basis functions are applied as linear additions to the baseline geometry as in the formulas
    below:
    .. math::
        y_{upper} = y_{upper}^{baseline} + \sum_i a_i f_i
        y_{lower} = y_{lower}^{baseline} + \sum_i b_i g_i
    
    The `a_i` and `b_i` above are the learnable parameters. There is a set for each basis family,
    that is, one set of `a_i`s for the polynomial-exponential basis functions and one for the
    sinusoidal basis functions. Similarly for `b_i`s.

    The basis functions are precomputed for efficiency and stored as buffers. The forward pass
    computes y-offsets by weighting the basis functions and adds them to the baseline coordinates.

    .. note::
        The implementation assumes that the x-coordinates lie in the range [0, 1].

    References:
        [1] Hicks, Raymond M., and Garret N. Vanderplaats. Application of numerical optimization to
            the design of supercritical airfoils without drag-creep. No. 770440. SAE Technical
            Paper, 1977.
        [2] Hicks, Raymond M., and Preston A. Henne. "Wing design by numerical optimization."
            Journal of aircraft 15.7 (1978): 407-412.
    """

    def __init__(
        self,
        X_upper_baseline: np.ndarray | torch.Tensor,
        X_lower_baseline: np.ndarray | torch.Tensor,
        polyexp_m: float,
        polyexp_n_list: list[int],
        sin_m: float,
        sin_n_list: list[int] | None = None,
        sin_n_count: int | None = None,
    ) -> None:
        r"""Initialize the Hicks-Henne representation.

        Args:
            X_upper_baseline: Upper surface baseline coordinates. Shape :math:`(N_upper, 2)`.
            X_lower_baseline: Lower surface baseline coordinates. Shape :math:`(N_lower, 2)`.
            polyexp_m: Scalar m for the polynomial-exponential basis family.
            polyexp_n_list: Exponent n list for the polynomial-exponential basis family.
            sin_m: Scalar m for the sinusoidal basis family.
            sin_n_list: Optional exponent n list for the sinusoidal basis family. Mutually exclusive
                with `sin_n_count`.
            sin_n_count: If provided, auto-generate this many sine bumps with even spaced peak
                locations in [0, 1] and convert to exponents n list.
        """
        super().__init__()

        if isinstance(X_upper_baseline, np.ndarray):
            X_upper_baseline = torch.from_numpy(X_upper_baseline)
        if isinstance(X_lower_baseline, np.ndarray):
            X_lower_baseline = torch.from_numpy(X_lower_baseline)

        # Check that x is between [0, 1] for the upper and lower surfaces
        if not torch.all((X_upper_baseline[:, 0] >= 0.0) & (X_upper_baseline[:, 0] <= 1.0)):
            raise ValueError('X_upper x-values must be in [0, 1].')
        if not torch.all((X_lower_baseline[:, 0] >= 0.0) & (X_lower_baseline[:, 0] <= 1.0)):
            raise ValueError('X_lower x-values must be in [0, 1].')
        
        
        # Save attributes in buffer so that they can be saved with state_dict
        self.register_buffer('X_upper_baseline', X_upper_baseline)
        self.register_buffer('X_lower_baseline', X_lower_baseline)
        # Convert exponents to tensors to store in register_buffer
        self.register_buffer('polyexp_m', torch.tensor(polyexp_m))
        self.register_buffer('polyexp_n_vals', torch.tensor(polyexp_n_list))
        self.register_buffer('sin_m', torch.tensor(sin_m))

        if sin_n_list is not None:
            self.register_buffer('sin_n_vals', torch.tensor(sin_n_list))
        else:
            if sin_n_count is None:
                raise ValueError('Provide either sin_n_list or sin_n_count.')
            x_peaks = torch.linspace(0.0, 1.0, sin_n_count)
            sin_n_list = _n_for_sin_peak_locations(x_peaks)
            self.register_buffer('sin_n_vals', torch.tensor(sin_n_list))
        

        # Learnable participation coefficients (K = count(n_vals) for each family)
        self.a_polyexp = nn.Parameter(torch.zeros(self.polyexp_n_vals.numel()))
        self.b_polyexp = nn.Parameter(torch.zeros(self.polyexp_n_vals.numel()))
        self.a_sin = nn.Parameter(torch.zeros(self.sin_n_vals.numel()))
        self.b_sin = nn.Parameter(torch.zeros(self.sin_n_vals.numel()))
        
        
        # Precompute the basis matrices to avoid repeated computation during forward()
        # This stores the basis values in a (N, K) matrix where
        # N - count(x), K - count(n_vals)
        self.register_buffer(
            'phi_upper_polyexp',
            _polyexp_basis(self.X_upper_baseline[:, 0], self.polyexp_n_vals, self.polyexp_m),
            persistent = False
        )
        self.register_buffer(
            'phi_lower_polyexp',
            _polyexp_basis(self.X_lower_baseline[:, 0], self.polyexp_n_vals, self.polyexp_m),
            persistent = False
        )
        self.register_buffer(
            'phi_upper_sin',
            _sin_basis(self.X_upper_baseline[:, 0], self.sin_n_vals, self.sin_m),
            persistent = False
        )
        self.register_buffer(
            'phi_lower_sin',
            _sin_basis(self.X_lower_baseline[:, 0], self.sin_n_vals, self.sin_m),
            persistent = False
        )


    def forward(self) -> tuple[torch.Tensor, torch.Tensor]:
        r"""Compute Hicks-Henne y-offsets for the upper and lower surfaces.

        Returns:
            tuple[torch.Tensor, torch.Tensor]: Upper and lower surface coordinates. `X_upper` of
                shape :math:`(N_upper, 2)` and `X_lower` of shape :math:`(N_lower, 2)`.
        """
        # Apply the bump function offsets to the upper and lower surfaces
        dy_upper = self.phi_upper_polyexp @ self.a_polyexp + self.phi_upper_sin @ self.a_sin
        dy_lower = self.phi_lower_polyexp @ self.b_polyexp + self.phi_lower_sin @ self.b_sin

        X_upper = torch.stack([self.X_upper_baseline[:, 0], self.X_upper_baseline[:, 1] + dy_upper],
                              dim = 1)
        X_lower = torch.stack([self.X_lower_baseline[:, 0], self.X_lower_baseline[:, 1] + dy_lower],
                              dim = 1)

        return X_upper, X_lower


    def visualize(self, ax = None):
        r"""Plot baseline and Hicks-Henne geometry.

        Args:
            ax: Optional Matplotlib Axes to draw on. If `None`, a new figure/axes is created.
        
        Returns:
            tuple[fig, ax]: The figure and axes used for plotting.
        """
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure

        # Move to CPU for matplotlib
        X_upper_baseline = self.X_upper_baseline.detach().cpu()
        X_lower_baseline = self.X_lower_baseline.detach().cpu()
        X_upper, X_lower = self.forward()
        X_upper, X_lower = X_upper.detach().cpu(), X_lower.detach().cpu()

        # Plot the baseline shape
        ax.plot(X_upper_baseline[:, 0], X_upper_baseline[:, 1], linestyle = '-', linewidth = 2,
                color = 'black', alpha = 0.7, label = 'upper (baseline)')
        ax.plot(X_lower_baseline[:, 0], X_lower_baseline[:, 1], linestyle = '--', linewidth = 2,
                color = 'black', alpha = 0.7, label = 'lower (baseline)')

        # Plot the shape
        ax.plot(X_upper[:, 0], X_upper[:, 1], linestyle = '-', linewidth = 2,
                color = 'orange', alpha = 0.7, label = 'upper')
        ax.plot(X_lower[:, 0], X_lower[:, 1], linestyle = '--', linewidth = 2,
                color = 'orange', alpha = 0.7, label = 'lower')

        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_aspect('equal')
        ax.set_title('Hicks-Henne Bump Function Parameterization')
        ax.legend(loc = 'upper center', bbox_to_anchor = (0.5, -0.12), ncol = 2)

        return fig, ax


import copy

import matplotlib.pyplot as plt
import numpy as np
import torch
from torch import nn

from geodiff.aux_nets import PreAuxNet


class NICE(nn.Module):
    r"""Nonlinear Independent Components Estimation (NICE) architecture as described in [1]_
    pre-composed with a Pre-Auxilliary network.

    This module generates a closed, non-self-intersecting base geometry via a PreAuxNet and then
    applies a stack of NICE layers.

    .. math:
        TODO

    References:
        [1]: Dinh, Laurent, David Krueger, and Yoshua Bengio. "Nice: Non-linear independent
            components estimation." arXiv preprint arXiv:1410.8516 (2014).
    """

    def __init__(
        self,
        geometry_dim: int,
        layer_count: int,
        coupling_net: nn.Module,
        preaux_net_layer_count: int,
        preaux_net_hidden_dim: int,
        preaux_net_act_f = nn.GELU,
        preaux_net_norm_f = nn.BatchNorm1d,
        preaux_net_out_f = nn.Softplus,
        volume_preserving: bool = False,
        use_batchnormalization: bool = False,
        use_residual_connection: bool = True,
    ) -> None:
        r"""Initialize the NICE object.

        Args:
            geometry_dim: Dimension of the output geometry, e.g. 2d, 3d etc.
            layer_count: Number of hidden layers in the NICE net.
            coupling_net: A torch network that is used as the coupling function.
            preaux_net_layer_count: Number of hidden layers in the Pre-Aux net. If `None` then only
                closed transform is performed.
            preaux_net_hidden_dim: Number of neurons in each hidden layer of the Pre-Aux net.
            preaux_net_act_f: Activation function used in the Pre-Aux net.
            preaux_net_norm_f: Normalization function used in the Pre-Aux net.
            preaux_net_out_f: Output activation function used in the Pre-Aux net.
            volume_preserving: If `False`, applies a learnable positive scaling at the output. If
                `True`, no scaling is applied and the transformation is volume preserving.
            use_batchnormalization: If `True`, applies batch normalization after each coupling
                layer.
            use_residual_connection: If `True`, applies a residual connection after each coupling
                layer.
        """
        super().__init__()

        # Save attributes in buffer so that they can be saved with state_dict
        self.register_buffer('geometry_dim', torch.tensor(geometry_dim, dtype = torch.int64))
        self.register_buffer('layer_count', torch.tensor(layer_count, dtype = torch.int64))
        if preaux_net_layer_count is not None:
            self.register_buffer('preaux_net_layer_count', torch.tensor(preaux_net_layer_count,
                                                                    dtype = torch.int64))
        else:
            self.register_buffer('preaux_net_layer_count', torch.tensor(-1, dtype = torch.int64))
        self.register_buffer('preaux_net_hidden_dim', torch.tensor(preaux_net_hidden_dim,
                                                                   dtype = torch.int64))
        self.register_buffer('volume_preserving', torch.tensor(volume_preserving,
                                                               dtype = torch.bool))
        self.register_buffer('use_batchnormalization', torch.tensor(use_batchnormalization,
                                                               dtype = torch.bool))
        self.register_buffer('use_residual_connection', torch.tensor(use_residual_connection,
                                                               dtype = torch.bool))


        # Create a Pre-Aux Net to close the shape
        self.preaux_net = PreAuxNet(
            geometry_dim = geometry_dim,
            layer_count = preaux_net_layer_count,
            hidden_dim = preaux_net_hidden_dim,
            act_f = preaux_net_act_f,
            norm_f = preaux_net_norm_f, 
            out_f = preaux_net_out_f
        )


        # Create a base mask to define the first partition
        d_partition = (geometry_dim // 2) if (geometry_dim % 2 == 0) else (geometry_dim + 1) // 2
        mask1 = torch.tensor([True] * d_partition + [False] * (geometry_dim - d_partition),
                                 dtype = torch.bool)
        mask2 = torch.tensor([False] * (geometry_dim - d_partition) + [True] * d_partition,
                                 dtype = torch.bool)
        self.register_buffer('mask1', mask1)
        self.register_buffer('mask2', mask2)


        # Create coupling nets for each NICE layer
        self.coupling_nets = nn.ModuleList()
        for i in range(layer_count):
            self.coupling_nets.append(copy.deepcopy(coupling_net))


        # Use batchnormalization with each coupling layer
        self.normalization_layers = nn.ModuleList()
        for i in range(layer_count):
            if use_batchnormalization:
                self.normalization_layers.append(nn.BatchNorm1d(geometry_dim))
            else:
                self.normalization_layers.append(nn.Identity())


        # Use positive diagonal scaling
        if not volume_preserving:
            self.log_scale = nn.Parameter(torch.zeros(geometry_dim))


    def forward(self, T: torch.Tensor = None, num_pts: int = 1000) -> torch.Tensor:
        r"""Compute the coordinates of the shape represented by the NICE object.

        Args:
            T: Options samples in the input domain which are mapped to points on the geometry.
                Mutually exclusive with `num_pts`.
            num_pts: Optional number of points :math:`N` to generate on the surface. Mututally
                exclusive with `T`.

        Returns:
            torch.Tensor: Matrix of coordinates of points on the geometry. Shape :math:`(N, d)`,
                where :math:`d` is the dimension of the geometry.
        """
        if T is None:
            if self.geometry_dim == 2:
                T = torch.linspace(0, 1, num_pts).reshape(-1, 1)
            elif self.geometry_dim == 3:
                t, s = torch.meshgrid(torch.linspace(0, 1, num_pts), torch.linspace(0, 1, num_pts),
                                      indexing = 'ij')
                T = torch.stack([t, s], dim = -1).reshape(-1, 2)
        # Put T on the same device as the model
        T.to(device = next(self.parameters()).device)

        # First create a closed transform using the PreAux net
        X = self.preaux_net(T)

        # Apply the NICE coupling layers
        mask_switch = True
        for coupling_net, normalization_layer in zip(self.coupling_nets, self.normalization_layers):
            # Pick a mask for the partition
            mask = self.mask1 if mask_switch else self.mask2

            if self.use_residual_connection:
                residual = X

            # Create the partition
            x1 = X[:, mask]
            x2 = X[:, ~mask]

            # Compute the output of the additive coupling layer
            y1 = x1
            y2 = x2 + coupling_net(x1)

            # Assign the y elements to the right place
            Y = torch.empty_like(X)
            Y[:, mask] = y1
            Y[:, ~mask] = y2

            # Assign Y to X to be used in the next layer as input
            X = Y

            # Apply batchnormalization
            X = normalization_layer(X)

            if self.use_residual_connection:
                # Apply residual connection
                X = (X + residual) / 2.0

            # Invert the mask for the next layer
            mask_switch = not mask_switch

        # Apply the final scaling if non-volume preserving
        if not self.volume_preserving:
            X = X * torch.exp(self.log_scale)

        return X


    def visualize(self, T: torch.Tensor = None, num_pts: int = 1000, ax = None):
        r"""Plot geometry represented by the NICE object.

        Args:
            T: Options samples in the input domain which are mapped to points on the geometry.
                Mutually exclusive with `num_pts`.
            num_pts: Optional number of points :math:`N` to generate on the surface. Mututally
                exclusive with `T`.
            ax: Optional Matplotlib Axes to draw on. If `None`, a new figure/axes is created.

        Returns:
            tuple[fig, ax]: The figure and axes used for plotting.
        """
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure

        # Move to CPU for matplotlib
        X = self.forward(T = T, num_pts = num_pts)
        X = X.detach().cpu()

        # Plot the shape
        ax.plot(X[:, 0], X[:, 1], linestyle = '-', linewidth = 2, color = 'orange', alpha = 0.7,
                label = 'curve')

        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_aspect('equal')
        ax.set_title('NICE parameterization')
        ax.legend(loc = 'upper center', bbox_to_anchor = (0.5, -0.12), ncol = 2)

        return fig, ax


import torch
from torch import nn


class MLP(nn.Module):
    r"""Flexible MLP: (Linear -> Norm -> Act) * layer_count, then Linear -> (optional out_act)
    """

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        layer_count: int,
        hidden_dim: int,
        act_f = nn.GELU,
        norm_f = nn.BatchNorm1d,
        out_f = None,
    ) -> None:
        r"""
        Args:
            input_dim: Dimension of input vector to the network.
            output_dim: Dimension of output vector from the network.
            layer_count: Number of hidden blocks.
            hidden_dim: Number of neurons in each hidden layer.
            act_f: Activation function, provided as class. Default: `nn.PReLU`.
            norm_f: Normalization function, provided as class. Default: `nn.BatchNorm1d`.
            out_f: Output layer activation function, provided as class. Default: None
                (`nn.Identity`).
        """
        super().__init__()

        if act_f is None:
            act_f = nn.Identity
        if norm_f is None:
            norm_f = nn.Identity
        if out_f is None:
            out_f = nn.Identity

        # Save attributes in buffer so that they can be saved with state_dict
        self.register_buffer('input_dim', torch.tensor(input_dim, dtype = torch.int64))
        self.register_buffer('output_dim', torch.tensor(output_dim, dtype = torch.int64))
        self.register_buffer('layer_count', torch.tensor(layer_count, dtype = torch.int64))
        self.register_buffer('hidden_dim', torch.tensor(hidden_dim, dtype = torch.int64))


        # Define the network blocks
        self.blocks = nn.ModuleList()

        # Get the first block
        self.blocks.append(nn.ModuleList([
            nn.Linear(input_dim, hidden_dim),
            norm_f(hidden_dim),
            act_f(),
        ]))
        # Add hidden blocks
        for i in range(layer_count - 1):
            self.blocks.append(nn.ModuleList([
                nn.Linear(hidden_dim, hidden_dim),
                norm_f(hidden_dim),
                act_f(),
            ]))
        # Add output block
        self.blocks.append(nn.ModuleList([
            nn.Linear(hidden_dim, output_dim),
            out_f(),
        ]))


    def forward(self, X):
        for block in self.blocks:
            for layer in block:
                X = layer(X)

        return X


class ResMLP(nn.Module):
    r"""Flexible Residual MLP: (Linear -> Norm -> Act + residual) / 2 * layer_count, then Linear ->
    (optional out_act).
    """

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        layer_count: int,
        hidden_dim: int,
        act_f = nn.GELU,
        norm_f = nn.BatchNorm1d,
        out_f = None,
    ) -> None:
        r"""
        Args:
            input_dim: Dimension of input vector to the network.
            output_dim: Dimension of output vector from the network.
            layer_count: Number of hidden blocks.
            hidden_dim: Number of neurons in each hidden layer.
            act_f: Activation function, provided as class. Default: `nn.PReLU`.
            norm_f: Normalization function, provided as class. Default: `nn.BatchNorm1d`.
            out_f: Output layer activation function, provided as class. Default: None
                (`nn.Identity`).
        """
        super().__init__()

        if act_f is None:
            act_f = nn.Identity
        if norm_f is None:
            norm_f = nn.Identity
        if out_f is None:
            out_f = nn.Identity

        # Save attributes in buffer so that they can be saved with state_dict
        self.register_buffer('input_dim', torch.tensor(input_dim, dtype = torch.int64))
        self.register_buffer('output_dim', torch.tensor(output_dim, dtype = torch.int64))
        self.register_buffer('layer_count', torch.tensor(layer_count, dtype = torch.int64))
        self.register_buffer('hidden_dim', torch.tensor(hidden_dim, dtype = torch.int64))


        # Define the network blocks
        self.blocks = nn.ModuleList()

        # Get the first block
        self.blocks.append(nn.ModuleList([
            nn.Linear(input_dim, hidden_dim),
            norm_f(hidden_dim),
            act_f(),
        ]))
        # Add hidden blocks
        for i in range(layer_count - 1):
            self.blocks.append(nn.ModuleList([
                nn.Linear(hidden_dim, hidden_dim),
                norm_f(hidden_dim),
                act_f(),
            ]))
        # Add output block
        self.blocks.append(nn.ModuleList([
            nn.Linear(hidden_dim, output_dim),
            out_f(),
        ]))


    def forward(self, X):
        for i, block in enumerate(self.blocks):
            # If for the first or last layer the dim does not match hidden dim, don't use residual
            # connection
            if i == 0:
                if self.input_dim != self.hidden_dim:
                    for layer in block:
                        X = layer(X)
                    continue
            elif i == len(self.blocks) - 1:
                if self.output_dim != self.hidden_dim:
                    for layer in block:
                        X = layer(X)
                    continue

            # In all other blocks use residual connection - this code is run if the input or output
            # dimension matches the hidden dimension
            residual = X
            for layer in block:
                X = layer(X)
            X = (X + residual) / 2.0

        return X


import math

import matplotlib.pyplot as plt
import numpy as np
import torch
from torch import nn


def _bernstein_basis(x: torch.Tensor, n: int) -> torch.Tensor:
    r"""Compute the Bernstein basis functions of degree :math:`n`.

    For each x value in `x`, k value in :math:`k = 0, \dots, n`, the Bernstein polynomial is:
    .. math::
        B_{k, n}(x) \;=\; \binom{n}{k}\, x^{k}\, \left(1 - x \right)^{n-k}.

    Args:
        x: x-coordinates of points to evaluate basis functions at. Shape :math:`(N,)`.
        n: Non-negative integer degree of the Bernstein polynomials. Produces :math:`n{+}1` columns.

    Returns:
        torch.Tensor: Basis matrix. Shape (N, K). Rows = x-samples, Cols = k-values.
    """
    dtype, device = x.dtype, x.device

    # Compute all binomial coefficients (n choose k) and exponents
    nCk = [math.comb(n, k) for k in range(n + 1)]
    nCk = torch.tensor(nCk, dtype = dtype, device = device)
    k = torch.arange(n + 1, dtype = dtype, device = device)

    # Reshape tensors to compute all basis functions at once
    # N - count(x), n + 1 - count(k), basis matrix returned (N, n + 1)
    # Make x (N, 1), k (1, n + 1) and use broadcasting to produce an output (N, n + 1)
    N = x.shape[0]
    x = x.view(N, 1)
    k = k.view(1, -1)
    nCk = nCk.view(1, -1)

    # Compute the basis
    B_kn = nCk * torch.pow(x, k) * torch.pow(1.0 - x, n - k)

    return B_kn


def _class_function(x: torch.Tensor, n1: float, n2: float) -> torch.Tensor:
    r"""Evaluate the CST *class* function :math:`C(x) = x^{n_1}(1-x)^{n_2}`.

    Args:
        x: x-coordinates of points to evaluate class function at. Shape :math:`(N,)`.
        n1: Class function exponent :math:`n_1`.
        n2: Class function exponent :math:`n_2`.

    Returns:
        torch.Tensor: :math:`C(x)` with the same shape and dtype as `x`.
    """
    return torch.pow(x, n1) * torch.pow(1 - x, n2)


class CST(nn.Module):
    r"""Class-Shape Transformation (CST) representation as described in [1]_.

    This module implements the CST representation as described in [1]_. The upper and lower surfaces
    of the shape are defined by:
    .. math::
        y_u(x) &= C(x)\,S_u(x) + \tau_u, \\
        y_l(x) &= C(x)\,S_l(x) + \tau_l,

    with shape functions
    .. math::
        S_u(x) &= \sum_{i=0}^{K_u-1} A_{u,i}\,S_i(x), \qquad
        S_l(x) = \sum_{i=0}^{K_l-1} A_{l,i}\,S_i(x),

    where the element shape functions :math:`S_i(x)` are the Bernstein polynomials of
    degree :math:`n = K-1`,
    .. math::
        S_i(x) \;=\; \binom{n}{i}\,x^{\,i}\,(1-x)^{\,n-i}, \qquad i = 0, \ldots, n.

    The class function is
    .. math::
        C(x) \;=\; x^{\,n_1}\,\bigl(1-x\bigr)^{\,n_2}.

    The scalars :math:`\tau_u` and :math:`\tau_l` are trailing-edge offsets that
    govern the trailing-edge thickness; specifically
    .. math::
        t_{\mathrm{TE}} \;=\; \tau_u - \tau_l .
    
    By default, all coefficients :math:`A_{u,i}` and :math:`A_{l,i}` are initialized to 1, so that
    :math:`\sum_i S_i(x) = 1` (partition of unity) and the initial surfaces reduce to
    :math:`y_u(x) = C(x) + \tau_u` and :math:`y_l(x) = C(x) + \tau_l`.

    References:
        [1]: Kulfan, Brenda, and John Bussoletti. "" Fundamental" parameteric geometry
            representations for aircraft component shapes." 11th AIAA/ISSMO multidisciplinary
            analysis and optimization conference. 2006.
    """

    def __init__(
        self,
        n1: float = 0.5,
        n2: float = 1.0,
        upper_basis_count: int = 9,
        lower_basis_count: int = 9,
        upper_te_thickness: float = 0.0,
        lower_te_thickness: float = 0.0,
    ) -> None:
        r"""Initialize the Hicks-Henne representation.

        Args:
            n1: Optional Class-function exponent :math:`n_1` in :math:`C(x) = x^{n_1}(1-x)^{n_2}`.
            n2: Optional Class-function exponent :math:`n_2` in :math:`C(x) = x^{n_1}(1-x)^{n_2}`.
            upper_basis_count: Optional Number of Bernstein basis functions for the upper surface.
                Internally uses Bernstein polynomial of degree ``n = upper_basis_count - 1``.
            lower_basis_count: Optional Number of Bernstein basis functions for the lower surface.
                Internally uses Bernstein polynomial of degree ``n = lower_basis_count - 1``.
            upper_te_thickness: Optional Trailing-edge offset added to the upper surface.
            lower_te_thickness: Optional Trailing-edge offset added to the lower surface.
        
        Raises:
            ValueError: If either of the upper or lower trailing edge thickness is negative.
        """
        super().__init__()

        if (upper_te_thickness < 0.0) or (lower_te_thickness < 0.0):
            raise ValueError('The upper and lower trailing edge thickness should be non-negative.')

        # Save attributes in buffer so that they can be saved with state_dict
        self.register_buffer('n1', torch.tensor(n1))
        self.register_buffer('n2', torch.tensor(n2))
        self.register_buffer('upper_basis_count', torch.tensor(upper_basis_count,
                                                               dtype = torch.int64))
        self.register_buffer('lower_basis_count', torch.tensor(lower_basis_count,
                                                               dtype = torch.int64))
        self.register_buffer('upper_te_thickness', torch.tensor(upper_te_thickness))
        self.register_buffer('lower_te_thickness', torch.tensor(lower_te_thickness))


        # Learnable coefficients for the Shape functions
        self.A_upper = nn.Parameter(torch.ones(self.upper_basis_count.item()))
        self.A_lower = nn.Parameter(-torch.ones(self.lower_basis_count.item()))


    def forward(
        self,
        x: torch.Tensor = None,
        num_pts: int = 100
    ) -> tuple[torch.Tensor, torch.Tensor]:
        r"""Compute the CST y-coordinates for upper and lower surfaces.

        Args:
            x: Optional abscissae. Shape :math:`(N,)`, values in :math:`[0, 1]`. Mutually exclusive
                with `num_pts`.
            num_pts: Optional number of points :math:`N` to generate on each surface. Mutually
                exclusive with `x`.

        Returns:
            tuple[torch.Tensor, torch.Tensor]: Upper and lower surface coordinates of shape
                :math:`(N, 2)`.
        
        Raises:
            ValueError: If any :math:`x \notin [0, 1]`.
        """
        
        # If x is provided check that it is in [0, 1] if not provided create equally spaced x
        if x is not None:
            if not torch.all((x >= 0.0) & (x <= 1.0)):
                raise ValueError('x-values must be in [0, 1].')
        else:
            dtype, device = self.n1.dtype, self.n1.device
            x = torch.linspace(0.0, 1.0, num_pts, dtype = dtype, device = device)
        
        # Get basis counts
        upper_basis_count = int(self.upper_basis_count.item())
        lower_basis_count = int(self.lower_basis_count.item())

        # Compute the class function and the Bernstein basis matrix
        C = _class_function(x, self.n1.item(), self.n2.item())
        B_upper = _bernstein_basis(x, upper_basis_count - 1)
        B_lower = _bernstein_basis(x, lower_basis_count - 1)

        # Compute the shape functions
        S_upper = B_upper @ self.A_upper
        S_lower = B_lower @ self.A_lower

        # Compute the y coordinates of points on the shape
        y_upper = C * S_upper + self.upper_te_thickness
        y_lower = C * S_lower - self.lower_te_thickness

        # Return the (x, y) coordinate pairs of points on the shape
        X_upper = torch.stack([x, y_upper], dim = 1)
        X_lower = torch.stack([x, y_lower], dim = 1)

        return X_upper, X_lower
    

    def visualize(self, x: torch.Tensor = None, num_pts: int = 100, ax = None):
        r"""Plot baseline and CST geometry.

        Args:
            x: Optional abscissae. Shape :math:`(N,)`, values in :math:`[0, 1]`. Mutually exclusive
                with `num_pts`.
            num_pts: Optional number of points :math:`N` to generate on each surface. Mutually
                exclusive with `x`.
            ax: Optional Matplotlib Axes to draw on. If `None`, a new figure/axes is created.
        
        Returns:
            tuple[fig, ax]: The figure and axes used for plotting.
        """
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure
        
        # Move to CPU for matplotlib
        X_upper, X_lower = self.forward(x = x, num_pts = num_pts)
        X_upper, X_lower = X_upper.detach().cpu(), X_lower.detach().cpu()

        # Plot the shape
        ax.plot(X_upper[:, 0], X_upper[:, 1], linestyle = '-', linewidth = 2,
                color = 'orange', alpha = 0.7, label = 'upper')
        ax.plot(X_lower[:, 0], X_lower[:, 1], linestyle = '--', linewidth = 2,
                color = 'orange', alpha = 0.7, label = 'lower')

        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_aspect('equal')
        ax.set_title('CST Parameterization')
        ax.legend(loc = 'upper center', bbox_to_anchor = (0.5, -0.12), ncol = 2)

        return fig, ax


import torch
import torch.nn as nn


class ChamferLoss(nn.Module):
    
    def __init__(self) -> None:
        super(ChamferLoss, self).__init__()
    
    
    def forward(self, Xc: torch.Tensor, Xt: torch.Tensor) -> torch.Tensor:
        # Compute pairwise squared Euclidean distances
        dists = torch.cdist(Xc, Xt, p = 2) ** 2  # (N, M)
        
        # Compute minimum distance for each point in Xc to Xt, and vice versa
        min_dist_Xc = torch.min(dists, dim = 1)[0]  # (N,)
        min_dist_Xt = torch.min(dists, dim = 0)[0]  # (M,)
        
        # Compute Chamfer distance
        loss = torch.mean(min_dist_Xc) + torch.mean(min_dist_Xt)
        return loss


import torch
from torch import nn

from geodiff.transforms import closed_transform_2d, closed_transform_3d
from geodiff.template_architectures import MLP, ResMLP


class PreAuxNet(nn.Module):
    r"""Pre-Auxilliary Network (Pre-Aux Net) for nD that creates an initial simple closed manifold.

    In :math:`nD` it first maps :math:`[0, 1]^n` to a :math:`(n+1)`-sphere and then transforms that
    to a simple closed manifold using anisotropic radial scaling.
    """

    def __init__(
        self,
        geometry_dim: int,
        layer_count: int,
        hidden_dim: int,
        act_f = nn.GELU,
        norm_f = nn.BatchNorm1d,
        out_f = nn.Softplus,
    ) -> None:
        r"""Initialize the PreAuxNet object.

        Args:
            geometry_dim: Dimension of the output geometry, e.g. 2d, 3d etc.
            layer_count: Number of hidden layers in the Pre-Aux net. If `None` then only closed
                transform is performed.
            hidden_dim: Number of neurons in each hidden layer.
            act_f: Activation function used in the hidden layers.
            norm_f: Normalization function used in the hidden layers.
            out_f: Output activation function used in the Pre-Aux net.
        """
        super().__init__()

        # Save attributes in buffer so that they can be saved with state_dict
        self.register_buffer('geometry_dim', torch.tensor(geometry_dim, dtype = torch.int64))
        if layer_count is not None:
            self.register_buffer('layer_count', torch.tensor(layer_count, dtype = torch.int64))
        else:
            self.register_buffer('layer_count', torch.tensor(-1, dtype = torch.int64))
        self.register_buffer('hidden_dim', torch.tensor(hidden_dim, dtype = torch.int64))


        # Close the input domain edges to map to the same r, e.g. theta = 0, 2pi in 2D
        if geometry_dim == 2:
            self.closed_transform = closed_transform_2d
        elif geometry_dim == 3:
            self.closed_transform = closed_transform_3d

        if layer_count is None:
            self.forward_stack = nn.Identity()
        else:
            # Get the network mapping input to r
            self.forward_stack = ResMLP(
                input_dim = geometry_dim,
                output_dim = 1,
                layer_count = layer_count,
                hidden_dim = hidden_dim,
                act_f = act_f,
                norm_f = norm_f,
                out_f = out_f
            )
    

    def forward(self, T: torch.Tensor) -> torch.Tensor:
        closed_manifold = self.closed_transform(T)
        r = self.forward_stack(closed_manifold)
        X = r * closed_manifold
        return X